from . import constant


class ExaError(Exception):
    """
    Generic PyEXASOL error, holds basic information about connection
    """
    def __init__(self, connection, message):
        self.connection = connection
        self.message = message

        super().__init__(self, message)

    def get_params_for_print(self):
        return {
            'message': self.message,
            'dsn': self.connection.options['dsn'],
            'user': self.connection.options['user'],
            'schema': self.connection.current_schema(),
            'session_id': self.connection.session_id(),
        }

    def __str__(self):
        if not self.connection.options['verbose_error']:
            return self.message

        params = self.get_params_for_print()
        pad_length = max(len(x) for x in params)
        res = ''

        for k in params:
            res += f"    {k.ljust(pad_length)}  =>  {params[k]}\n"

        return '\n(\n' + res + ')\n'


class ExaRuntimeError(ExaError):
    pass


class ExaCommunicationError(ExaError):
    """
    WebSocket communication failure after connection was established
    """
    pass


class ExaRequestError(ExaError):
    """
    Generic error returned from Exasol server after making a request
    """
    def __init__(self, connection, code, message):
        self.code = code

        super().__init__(connection, message)

    def get_params_for_print(self):
        params = super().get_params_for_print()
        params['code'] = self.code

        return params


class ExaAuthError(ExaRequestError):
    """
    Connection was established successfully, but authorization failed
    """
    pass


class ExaQueryError(ExaRequestError):
    """
    Error returned from Exasol server specifically for SQL query request (EXECUTE)
    """
    def __init__(self, connection, query, code, message):
        self.query = query

        super().__init__(connection, code, message)

    def get_params_for_print(self):
        params = super().get_params_for_print()
        params['session_id'] = self.connection.session_id()

        if len(self.query) > constant.EXCEPTION_QUERY_TEXT_MAX_LENGTH:
            params['query'] = f'{self.query[:constant.EXCEPTION_QUERY_TEXT_MAX_LENGTH]}' \
                              f'\n------ TRUNCATED TOO LONG QUERY ------\n'
        else:
            params['query'] = self.query

        return params


class ExaQueryTimeoutError(ExaQueryError):
    """
    Specific error for SQL query reaching QUERY_TIMEOUT and being terminated by server
    """
    pass


class ExaQueryAbortError(ExaQueryError):
    """
    Specific error for SQL query being aborted with .abort_query() call or KILL STATEMENT
    """
    pass


class ExaConnectionError(ExaError):
    """
    Generic error for connection failures
    """
    pass


class ExaConnectionDsnError(ExaConnectionError):
    """
    Specific error for connection failure related to DSN (connection string) issues
    """
    pass


class ExaConnectionFailedError(ExaConnectionError):
    """
    Specific error related to establishing WebSocket communication
    """
    pass


class ExaConcurrencyError(ExaError):
    """
    Detected an attempt to run multiple queries in multiple threads at the same time
    """
    pass



import string
import re


class ExaFormatter(string.Formatter):
    safe_ident_regexp = re.compile(r'^[A-Za-z_][A-Za-z0-9_]*$')
    safe_decimal_regexp = re.compile(r'^(\+|-)?[0-9]+(\.[0-9]+)?$')
    safe_float_regexp = re.compile(r'^(\+|-)?[0-9]+(\.[0-9]+((e|E)(\+|-)[0-9]+)?)?$')

    def __init__(self, connection):
        self.connection = connection

        self.conversions = {
            's': self.quote,
            'd': self.safe_decimal,
            'f': self.safe_float,
            'i': self.safe_ident,
            'q': self.quote_ident,
            'r': str
        }

        self.default_conversion = 's'

        # Set default treatment for identifiers passed as strings to relevant functions
        if self.connection.options['quote_ident']:
            self.default_format_ident = self.quote_ident  # Identifiers will be quoted and escaped
            self.default_format_ident_value = str         # Identifier values will be left unchanged
        else:
            self.default_format_ident = self.safe_ident   # Identifiers will only be checked for safety
            self.default_format_ident_value = str.upper   # Identifier values will be transformed to upper-case

    def format_field(self, value, format_spec):
        if format_spec != '':
            raise ValueError("format_spec is disabled for ExaFormatter")

        return value

    def convert_field(self, value, conversion):
        if conversion is None:
            conversion = self.default_conversion

        if conversion not in self.conversions:
            raise ValueError(f"Unknown conversion {conversion}")

        if isinstance(value, list):
            if not value:
                raise ValueError("Trying to format an empty list")
            return ', '.join([self.conversions[conversion](v) for v in value])
        else:
            return self.conversions[conversion](value)

    @classmethod
    def escape(cls, val):
        return str(val).replace("'", "''")

    @classmethod
    def escape_ident(cls, val):
        return str(val).replace('"', '""')

    @classmethod
    def escape_like(cls, val):
        return cls.escape(val).replace('\\', '\\\\').replace('%', r'\%').replace('_', r'\_')

    @classmethod
    def quote(cls, val):
        if val is None:
            return 'NULL'

        return f"'{cls.escape(val)}'"

    @classmethod
    def quote_ident(cls, val):
        if isinstance(val, tuple):
            return '.'.join([cls.quote_ident(x) for x in val])

        return f'"{cls.escape_ident(val)}"'

    @classmethod
    def safe_ident(cls, val):
        if isinstance(val, tuple):
            return '.'.join([cls.safe_ident(x) for x in val])

        val = str(val)

        if not cls.safe_ident_regexp.match(val):
            if '.' in val:
                parts = val.split('.')
                raise ValueError(f"Value [{val}] is not a safe identifier. Please use tuple to pass schema names. "
                                 f"Example: ('{parts[0]}', '{parts[1]}')")
            elif '"' in val:
                raise ValueError(f"Value [{val}] is not a safe identifier. Please use 'quote_ident' or '!q' conversion "
                                 f"to pass identifiers with lowercase or special characters.")
            else:
                raise ValueError(f"Value [{val}] is not a safe identifier")

        return val

    @classmethod
    def safe_float(cls, val):
        if val is None:
            return 'NULL'

        val = str(val)

        if not cls.safe_float_regexp.match(val):
            raise ValueError(f'Value [{val}] is not a safe float')

        return val

    @classmethod
    def safe_decimal(cls, val):
        if val is None:
            return 'NULL'

        val = str(val)

        if not cls.safe_decimal_regexp.match(val):
            raise ValueError(f'Value [{val}] is not a safe integer')

        return val

    def __repr__(self):
        return f'<{self.__class__.__name__} session_id={self.connection.session_id()}>'



import os
import sys
import subprocess


class ExaScriptOutputProcess(object):
    def __init__(self, host, port, output_dir):
        self.host = host
        self.port = port

        self.output_dir = output_dir
        self.output_address = None

        self.proc = None

    def start(self):
        args = [sys.executable,
                '-m', 'pyexasol_utils.script_output',
                '--output-dir', str(self.output_dir),
                '--ppid', str(os.getpid())
                ]

        if self.host:
            args.append('--host')
            args.append(self.host)

        if self.port:
            args.append('--port')
            args.append(str(self.port))

        self.proc = subprocess.Popen(args, stdout=subprocess.PIPE)
        self.output_address = self.proc.stdout.readline().decode().rstrip('\n')

        self.proc.stdout.close()

    def get_output_address(self):
        if self.output_address is None:
            raise RuntimeError("Script output address 'host:port' is not available")

        return self.output_address

    def join(self):
        if self.proc:
            return self.proc.wait()

        return None

    def join_with_exc(self):
        code = self.join()

        if code != 0:
            raise RuntimeError(f"Script output server process finished with exitcode: {code}")

        return code

    def terminate(self):
        if self.proc:
            self.proc.terminate()



PROTOCOL_V1 = 1
PROTOCOL_V2 = 2

DEFAULT_PORT = 8563
DEFAULT_AUTOCOMMIT = True

DEFAULT_CONNECTION_TIMEOUT = 10
DEFAULT_SOCKET_TIMEOUT = 30
DEFAULT_QUERY_TIMEOUT = 0

DEFAULT_FETCHMANY_SIZE = 10000
DEFAULT_FETCH_SIZE_BYTES = 5 * 1024 * 1024

DRIVER_NAME = 'PyEXASOL'

LOGGER_FILENAME_TIMESTAMP_FORMAT = '%Y%m%d_%H%M%S_%f'
LOGGER_MAX_JSON_LENGTH = 20000

EXCEPTION_QUERY_TEXT_MAX_LENGTH = 20000



"""
DB-API 2 compatibility package
https://www.python.org/dev/peps/pep-0249/

It may be used for a short time to test PyEXASOL with existing applications
It is highly recommended to switch to native interface before any production usage

There is no "paramstyle" and no proper error handling
"""

from ..connection import ExaConnection

apilevel = '2.0'
threadsafety = 1
paramstyle = None


def connect(**kwargs):
    if 'autocommit' not in kwargs:
        kwargs['autocommit'] = False

    return DB2Connection(**kwargs)


class DB2Connection(ExaConnection):
    def cursor(self):
        return DB2Cursor(self)


class DB2Cursor(object):
    arraysize = 1

    def __init__(self, connection):
        self.connection = connection
        self.stmt = None

    def execute(self, query):
        self.stmt = self.connection.execute(query)

    def executemany(self, query):
        raise NotSupportedError

    def fetchone(self):
        return self.stmt.fetchone()

    def fetchmany(self, size=None):
        if size is None:
            size = self.arraysize

        return self.stmt.fetchmany(size)

    def fetchall(self):
        return self.stmt.fetchall()

    def nextset(self):
        raise NotSupportedError

    def setinputsizes(self):
        pass

    def setoutputsize(self):
        pass

    @property
    def description(self):
        cols = []

        for k, v in self.stmt.columns().items():
            cols.append((
                k,
                v.get('type', None),
                v.get('size', None),
                v.get('size', None),
                v.get('precision', None),
                v.get('scale', None),
                True
            ))

        return cols

    @property
    def rowcount(self):
        return self.stmt.rowcount()

    def close(self):
        self.stmt.close()


class NotSupportedError(Exception):
    pass



import decimal
import datetime


class ExaTimeDelta(datetime.timedelta):
    def reverse_seconds(self):
        if self.microseconds > 0:
            seconds = 86399 - self.seconds
            microseconds = 1000000 - self.microseconds
        elif self.seconds > 0:
            seconds = 86400 - self.seconds
            microseconds = 0
        else:
            seconds = 0
            microseconds = 0
        return seconds, microseconds

    @classmethod
    def from_interval(cls, val):
        td = cls(
            days=int(val[0:10]),
            hours=int(val[11:13]),
            minutes=int(val[14:16]),
            seconds=int(val[17:19]),
            microseconds=int(round(float(val[20:29].ljust(9, '0')) / 1000)) if len(val) > 20 else 0,
        )
        if val[0] == '-':
            # normalize according to Python timedelta rules (days are negative; remaining parts apply back "up" towards 0)
            # - eg. -6 days, 1:00:00.000000 would represent 5 days, 23 hours ago (6 days back, 1 hour forward)
            (seconds, microseconds) = td.reverse_seconds()
            if seconds or microseconds:
                td = cls(days=td.days - 1, seconds=seconds, microseconds=microseconds)
            else:
                td = cls(days=td.days, seconds=seconds, microseconds=microseconds)
        return td

    @classmethod
    def from_timedelta(cls, td):
        return cls(days=td.days, seconds=td.seconds, microseconds=td.microseconds)

    def to_interval(self):
        if self.days < 0:
            (seconds, microseconds) = self.reverse_seconds()
            if seconds or microseconds:
                s = "-%09d " % (abs(self.days) - 1,)
            else:
                s = "-%09d " % (abs(self.days) - 1,)
        else:
            seconds = self.seconds
            microseconds = self.microseconds
            s = "+%09d " % (self.days,)

        mm, ss = divmod(seconds, 60)
        hh, mm = divmod(mm, 60)
        s = s + "%02d:%02d:%02d.%09d" % (hh, mm, ss, microseconds * 1000)
        return s

    def __str__(self):
        return self.to_interval()


def exasol_mapper(val, data_type):
    """
    Convert into Python 3 data types according to Exasol manual

    strptime() function is slow, so we use direct string slicing for performance sake
    More details about this problem: http://ze.phyr.us/faster-strptime/

    DECIMAL(p,0)           -> int
    DECIMAL(p,s)           -> decimal.Decimal
    DOUBLE                 -> float
    DATE                   -> datetime.date
    TIMESTAMP              -> datetime.datetime
    BOOLEAN                -> bool
    VARCHAR                -> str
    CHAR                   -> str
    INTERVAL DAY TO SECOND -> datetime.timedelta
    <others>               -> str
    """

    if val is None:
        return None
    elif data_type['type'] == 'DECIMAL':
        if data_type['scale'] == 0:
            return int(val)
        else:
            return decimal.Decimal(val)
    elif data_type['type'] == 'DATE':
        return datetime.date(int(val[0:4]), int(val[5:7]), int(val[8:10]))
    elif data_type['type'] == 'TIMESTAMP':
        return datetime.datetime(int(val[0:4]), int(val[5:7]), int(val[8:10]),           # year, month, day
                                 int(val[11:13]), int(val[14:16]), int(val[17:19]),      # hour, minute, second
                                 int(val[20:26].ljust(6, '0')) if len(val) > 20 else 0)  # microseconds (if available)
    elif data_type['type'] == 'INTERVAL DAY TO SECOND':
        return ExaTimeDelta.from_interval(val)
    else:
        return val



from setuptools import setup
import pathlib

__version__ = None
version_path = pathlib.Path(__file__).parent / 'pyexasol/version.py'
exec(version_path.read_text())

setup(
    name='pyexasol',
    version=__version__,
    description='Exasol python driver with extra features',
    long_description="""
Exasol driver with low overhead, fast HTTP transport and compression. It is implemented in Python 3.6+.

Main features:

-  Based on WebSocket protocol;
-  Optimized for minimum overhead;
-  Easy integration with pandas via HTTP transport;
-  Compression to reduce network bottleneck;

PyEXASOL does not follow DB-API 2.0 in favor of Exasol-specific features.

Please read "Best practices" manual page to learn how to use PyEXASOL with maximum efficiency.
    """,
    long_description_content_type='text/markdown',
    url='https://github.com/badoo/pyexasol',
    author='Vitaly Markov',
    author_email='wild.desu@gmail.com',

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Database',
    ],

    keywords='exasol sql database performance websocket import export',

    packages=[
        'pyexasol',
        'pyexasol.db2',
        'pyexasol_utils',
    ],

    install_requires=[
        'websocket-client>=0.47.0',
        'rsa',
    ],

    extras_require={
        'pandas': ['pandas>=0.22,!=0.23.1'],
        'encrypt': ['pyopenssl>=17.5.0'],
        'ujson': ['ujson'],
        'rapidjson': ['python-rapidjson'],
        'examples': ['pproxy', 'psutil'],
    },

    python_requires='>=3.6',
)
